# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes defining OT problem(s) (objective function + utilities)."""

from typing import Callable, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np
from ott.core import problems
from ott.core import sinkhorn_lr
from ott.geometry import epsilon_scheduler
from ott.geometry import geometry
from ott.geometry import low_rank
from ott.geometry import pointcloud
# Because Protocol is not available in Python < 3.8
from typing_extensions import Protocol


class Transport(Protocol):
  """Defines the interface for the solution of a transport problem.

  Classes implementing those function do not have to inherit from it, the
  class can however be used in type hints to support duck typing.
  """

  @property
  def matrix(self) -> jnp.ndarray:
    ...

  def apply(self, inputs: jnp.ndarray, axis: int) -> jnp.ndarray:
    ...

  def marginal(self, axis: int = 0) -> jnp.ndarray:
    ...


LossTerm = Callable[[jnp.ndarray], jnp.ndarray]
Loss = Tuple[Tuple[LossTerm, LossTerm], Tuple[LossTerm, LossTerm]]


def make_square_loss():
  return ((lambda x: x**2, lambda y: y**2), (lambda x: x,
                                             lambda y: 2.0 * y))


def make_kl_loss(clipping_value: float = 1e-8):

  return (
      (lambda x: -jax.scipy.special.entr(x) - x, lambda y: y),
      (lambda x: x, lambda y: jnp.log(jnp.clip(y, clipping_value)))
  )


@jax.tree_util.register_pytree_node_class
class QuadraticProblem:
  """Holds the definition of the quadratic regularized OT problem.

  The quadratic loss of a single OT matrix is assumed to
  have the form given in Eq. 4 from

  http://proceedings.mlr.press/v48/peyre16.pdf

  The two geometries below parameterize matrices C and bar{C} in that equation.
  The function L (of two real values) in that equation is assumed
  to match the form given in Eq. 5. , with our notations:

  L(x, y) = lin1(x) + lin2(y) - quad1(x) * quad2(y)
  """

  def __init__(self,
               geom_xx: geometry.Geometry,
               geom_yy: geometry.Geometry,
               geom_xy: Optional[geometry.Geometry] = None,
               fused_penalty: Optional[float] = None,
               a: Optional[jnp.ndarray] = None,
               b: Optional[jnp.ndarray] = None,
               loss: Optional[Loss] = None,
               tau_a: Optional[float] = 1.0,
               tau_b: Optional[float] = 1.0,
               gw_unbalanced_correction: Optional[bool] = True):
    """Initializes the QuadraticProblem.

    Args:
      geom_xx: the geometry.Geometry object defining the ground geometry / cost
        of the first space.
      geom_yy: the geometry.Geometry object defining the ground geometry / cost
        of the second space.
      geom_xy: the geometry.Geometry object defining the linear penalty term
        for Fused Gromov Wasserstein. If None, the problem reduces to a plain
        Gromov Wasserstein problem.
      fused_penalty: multiplier of the linear term in Fused Gromov Wasserstein,
        i.e. problem = purely quadratic + fused_penalty * linear problem. If
        fused_penalty is None but geom_xy is passed, fused_penalty is set by
        default to 1.0, equal to 0.0 otherwise.
      a: jnp.ndarray[n] representing the probability weights of the samples
        from geom_xx. If None, it will be uniform.
      b: jnp.ndarray[n] representing the probability weights of the samples
        from geom_yy. If None, it will be uniform.
      loss: a 2-tuple of 2-tuples of Callable. The first tuple is the linear
        part of the loss (see in the pydoc of the class lin1, lin2). The second
        one is the quadratic part (quad1, quad2). If None is passed, the loss
        is set as the 4 functions representing the squared euclidean loss, and
        this property is taken advantage of in subsequent computations. See
        make_kl_loss for an alternative, no less optimized way of setting the
        loss.
      tau_a: if lower that 1.0, defines how much unbalanced the problem is on
        the first marginal.
      tau_b: if lower that 1.0, defines how much unbalanced the problem is on
        the second marginal.
      gw_unbalanced_correction: True (default) if the unbalanced version of
        Sejourne et al. (Neurips 2021) is used, False if tau_a and tau_b
        only affect the inner Sinhkorn loop.
    """
    self.geom_xx = geom_xx
    self.geom_yy = geom_yy
    self.geom_xy = geom_xy
    if fused_penalty is None:
      fused_penalty = jnp.where(self.geom_xy is None, 0.0, 1.0)
    self.fused_penalty = fused_penalty
    self._a = a
    self._b = b
    self.tau_a = tau_a
    self.tau_b = tau_b
    self.gw_unbalanced_correction = gw_unbalanced_correction
    if loss is None:
      self._sq_euc = True
      self.loss = make_square_loss()
    else:
      self._sq_euc = False
      self.loss = loss

  @property
  def is_fused(self):
    return self.geom_xy is not None and jnp.all(self.fused_penalty > 0.0)

  @property
  def is_all_geoms_lr(self):
    lr_geoms = (
        isinstance(self.geom_xx, low_rank.LRCGeometry) and
        isinstance(self.geom_yy, low_rank.LRCGeometry))
    lr_geoms = lr_geoms and (
        isinstance(self.geom_xy, low_rank.LRCGeometry)
        or
        not self.is_fused
        )

  @property
  def linear_loss(self):
    return self.loss[0]

  @property
  def quad_loss(self):
    return self.loss[1]

  @property
  def is_balanced(self):
    return ((not self.gw_unbalanced_correction)
            or (self.tau_a == 1.0 and self.tau_b == 1.0))

  def tree_flatten(self):
    return ([
        self.geom_xx, self.geom_yy, self.geom_xy, self.fused_penalty, self._a,
        self._b
    ],
            {'tau_a': self.tau_a, 'tau_b': self.tau_b, 'loss': self.loss,
             'gw_unbalanced_correction': self.gw_unbalanced_correction}
            )

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    return cls(*children, **aux_data)

  @property
  def a(self):
    num_a = self.geom_xx.shape[0]
    return jnp.ones((num_a,)) / num_a if self._a is None else self._a

  @property
  def b(self):
    num_b = self.geom_yy.shape[0]
    return jnp.ones((num_b,)) / num_b if self._b is None else self._b

  def marginal_dependent_cost(self, marginal_1, marginal_2):
    r"""Initialises cost term that depends on the marginals of the transport.

    Uses the first term in Equation 6, Proposition 1 of
    http://proceedings.mlr.press/v48/peyre16.pdf.

    Let :math:`p` [num_a,] be the marginal of the transport matrix for samples
    from `geom_xx` and :math:`q` [num_b,] be the marginal of the transport
    matrix for samples from `geom_yy`. `cost_xx` (resp. `cost_yy`) is the
    cost matrix of `geom_xx` (resp. `geom_yy`). The cost term that
    depends on these marginals can be written as:

    `marginal_dep_term` = `lin1`(`cost_xx`) :math:`p \mathbb{1}_{num_b}^T`
                      + (`lin2`(`cost_yy`) :math:`q \mathbb{1}_{num_a}^T)^T`

    Args:
      marginal_1: jnp.ndarray<float>[num_a,], marginal of the transport matrix
       for samples from geom_xx
      marginal_2: jnp.ndarray<float>[num_b,], marginal of the transport matrix
       for samples from geom_yy

    Returns:
      a LRCGeometry.
    """
    if self._sq_euc:  # quadratic apply
      tmp1 = self.geom_xx.apply_square_cost(marginal_1, axis=1)
      tmp2 = self.geom_yy.apply_square_cost(marginal_2, axis=1)
    else:
      tmp1 = self.geom_xx.apply_cost(marginal_1, axis=1,
                                     fn=self.linear_loss[0])
      tmp2 = self.geom_yy.apply_cost(marginal_2, axis=1,
                                     fn=self.linear_loss[1])
    x_term = jnp.concatenate((tmp1, jnp.ones_like(tmp1)), axis=1)
    y_term = jnp.concatenate((jnp.ones_like(tmp2), tmp2), axis=1)
    return low_rank.LRCGeometry(cost_1=x_term, cost_2=y_term)

  def cost_unbalanced_correction(self, transport_matrix, marginal_1, marginal_2,
                                 epsilon, rescale_factor, delta=1e-9) -> float:
    r"""Calculates cost term from the quadratic divergence when unbalanced.

    In the unbalanced setting (i.e. tau_a<1.0 or tau_b<1.0), the
    introduction of a quadratic divergence (see Sejourne et al. Neurips 2021)
    adds a term to the GW local cost.

    Let :math:`a` [num_a,] be the target weights for samples
    from geom_xx and :math:`b` [num_b,] be the target weights
    for samples from `geom_yy`. Let :math:`P` [num_a, num_b] be the transport
    matrix, :math:`P1` the first marginal and :math:`P^T1` the second marginal.
    The term of the cost matrix coming from the quadratic KL in the
    unbalanced case can be written as:

    `unbalanced_correction_term` =
        :math:`tau_a / (1 - tau_a) * \sum(KL(P1|a))`
        :math:`+ tau_b / (1 - tau_b) * \sum(KL(P^T1|b))`
        :math:`+ epsilon * \sum(KL(P|ab'))`

    Arguments:
      transport_matrix: jnp.ndarray<float>[num_a, num_b], transport matrix.
      marginal_1: jnp.ndarray<float>[num_a,], marginal of the transport matrix
       for samples from geom_xx
      marginal_2: jnp.ndarray<float>[num_b,], marginal of the transport matrix
       for samples from geom_yy
      epsilon: regulariser.
      rescale_factor: float, scaling factor for the transport matrix.
      delta: float, small quantity to avoid diverging KLs.

    Returns:
      float, cost term
    """
    def regulariser(tau):
      return tau / (1.0 - tau) if tau != 1.0 else 0

    cost = regulariser(self.tau_a) * jax.scipy.special.xlogy(
        marginal_1,
        rescale_factor * marginal_1 / jnp.clip(self.a, a_min=delta)).sum()
    cost += regulariser(self.tau_b) * jax.scipy.special.xlogy(
        marginal_2,
        rescale_factor * marginal_2 / jnp.clip(self.b, a_min=delta)).sum()
    cost += epsilon * jax.scipy.special.xlogy(
        transport_matrix,
        rescale_factor * transport_matrix
        / jnp.clip(self.a[:, None] * self.b[None, :], a_min=delta)).sum()
    return cost

  def init_transport(self):
    # TODO(oliviert, cuturi): consider passing a custom initialization.
    a = jax.lax.stop_gradient(self.a)
    b = jax.lax.stop_gradient(self.b)
    return (a[:, None] * b[None, :] if self.is_balanced else a[:, None] *
            b[None, :] / jnp.sqrt(a.sum() * b.sum()))

  def init_transport_mass(self) -> float:
    """Initialises the transport mass.

    Returns:
      A float, sum of the elements of the normalised transport matrix.
    """
    a = jax.lax.stop_gradient(self.a)
    b = jax.lax.stop_gradient(self.b)
    transport_mass = a.sum() * b.sum()
    return (transport_mass if self.is_balanced
            else transport_mass / jnp.sqrt(transport_mass))

  def init_linearization(
      self,
      epsilon: Optional[Union[epsilon_scheduler.Epsilon, float]] = None
  ) -> problems.LinearProblem:
    """Initialises a linear problem locally around a naive initializer ab'.

    If the problem is balanced (`tau_a=1.0 and tau_b=1.0'), the equation of the
    cost follows Equation 6, Proposition 1 of
    http://proceedings.mlr.press/v48/peyre16.pdf.
    If the problem is unbalanced (`tau_a<1.0 or tau_b<1.0`), there are two
    possible cases. A first possibility is to introduce a
    quadratic KL divergence on the marginals in the objective as done
    in Sejourne et al. 2021 (https://arxiv.org/abs/2009.04266)
    (`gw_unbalanced_correction=True`), which in turns modifies the local cost
    matrix.
    Alternatively, it could be possible to leave
    the formulation of the local cost unchanged, i.e. follow Equation 6 of
    Proposition 1 (`gw_unbalanced_correction=False`) and include the
    unbalanced terms at the level of the linear problem only.

    Let :math:`P` [num_a, num_b] be the transport matrix, `cost_xx` is the
    cost matrix of `geom_xx` and `cost_yy` is the cost matrix of `geom_yy`.
    `left_x` and `right_y` depend on the loss chosen for GW.
    `gw_unbalanced_correction` is an boolean indicating whether or not the
    unbalanced correction applies.
    The equation of the local cost can be written as:

    `cost_matrix` = `marginal_dep_term`
                + `left_x`(`cost_xx`) :math:`P` `right_y`(`cost_yy`):math:`^T`
                + `unbalanced_correction` * `gw_unbalanced_correction`

    When working with the fused problem, a linear term is added to the cost
    matrix:
    `cost_matrix` += `fused_penalty` * `geom_xy.cost_matrix`


    Args:
      epsilon: An epsilon scheduler or a float passed on to the linearization.

    Returns:
      A problems.LinearProblem, representing local linearization of GW problem.
    """
    unbalanced_correction = 0.0
    tmp = self.init_transport()
    marginal_1 = tmp.sum(1)
    marginal_2 = tmp.sum(0)

    # Initialises cost.
    marginal_cost = self.marginal_dependent_cost(marginal_1, marginal_2)

    if not self.is_balanced:
      unbalanced_correction = self.cost_unbalanced_correction(
          tmp, marginal_1, marginal_2, epsilon, 1.0)

    tmp = self.geom_xx.apply_cost(tmp, axis=1, fn=self.quad_loss[0])
    tmp = self.geom_yy.apply_cost(tmp.T, axis=1, fn=self.quad_loss[1]).T
    cost_matrix = (marginal_cost.cost_matrix - tmp + unbalanced_correction)

    # Initialises epsilon for Unbalanced GW according to Sejourne et al (2021).
    if not self.is_balanced:
      transport_mass = marginal_1.sum()
      epsilon = update_epsilon_unbalanced(epsilon, transport_mass)

    cost_matrix += self.fused_penalty * jnp.where(
        self.is_fused,
        0.0 if self.geom_xy is None else self.geom_xy.cost_matrix,
        0.0)

    geom = geometry.Geometry(cost_matrix=cost_matrix, epsilon=epsilon)
    return problems.LinearProblem(
        geom, self.a, self.b, tau_a=self.tau_a, tau_b=self.tau_b)

  def init_lr_linearization(
      self,
      rank: int = 10,
      **kwargs
  ) -> problems.LinearProblem:
    """Linearizes a Quad problem with a predefined initializer."""
    x_ = self.geom_xx.apply_square_cost(self.a)
    y_ = self.geom_yy.apply_square_cost(self.b)
    geom_ = pointcloud.PointCloud(x_, y_).to_LRCGeometry()
    out = sinkhorn_lr.LRSinkhorn(rank=rank, **kwargs)(
        problems.LinearProblem(geom_, self.a, self.b))
    return problems.LinearProblem(
        self.update_lr_geom(out),
        self.a, self.b, tau_a=self.tau_a, tau_b=self.tau_b)

  def update_lr_geom(self, lr_sink):
    """Using LR Sinkhorn output, recompute (possibly LRC) linearization."""
    marginal_1 = lr_sink.marginal(1)
    marginal_2 = lr_sink.marginal(0)
    marginal_cost = self.marginal_dependent_cost(marginal_1, marginal_2)

    # Extract factors from LR Sinkhorn output
    q, r, inv_sqg = lr_sink.q, lr_sink.r, 1.0 / jnp.sqrt(lr_sink.g)
    # Distribute middle marginal evenly across both factors.
    q, r = q * inv_sqg[None, :], r * inv_sqg[None, :]

    # Handle LRC Geometry case.
    if self.is_all_geoms_lr:
      tmp1r = self.geom_xx.apply_cost_2(q)
      tmp2l = jnp.transpose(self.geom_yy.apply_cost_1(r, 1))
      geom = low_rank.LRCGeometry(cost_1=tmp1r, cost_2=-tmp2l)
      geom = low_rank.add_lrc_geom(geom, marginal_cost)
      if self.is_fused:
        geom = low_rank.add_lrc_geom(
            geom, self.geom_xy)

    else:
      tmp1 = self.geom_xx.apply_cost(q, axis=1, fn=self.quad_loss[0])
      tmp2 = self.geom_yy.apply_cost(r, axis=1, fn=self.quad_loss[1])
      cost_matrix = marginal_cost.cost_matrix - jnp.dot(tmp1, tmp2.T)

      cost_matrix += self.fused_penalty * jnp.where(
          self.is_fused,
          0.0 if self.geom_xy is None else self.geom_xy.cost_matrix, 0.0)
      geom = geometry.Geometry(cost_matrix=cost_matrix)
    return geom

  def update_linearization(
      self,
      transport: Transport,
      epsilon: Optional[Union[epsilon_scheduler.Epsilon, float]] = None,
      old_transport_mass: float = 1.0) -> problems.LinearProblem:
    """Updates linearization of GW problem by updating cost matrix.

    If the problem is balanced (`tau_a=1.0 and tau_b=1.0`), the equation
    follows Equation 6, Proposition 1 of
    http://proceedings.mlr.press/v48/peyre16.pdf. If the problem is unbalanced
    (`tau_a<1.0 or tau_b<1.0`), two cases are possible, as explained in the
    pydoc of `init_linearization` above. Finally, it is also possible to
    consider a Fused Gromov Wasserstein problem. Details about the resulting
    cost matrix are given in the pydoc of `init_linearization`.

    Args:
      transport: Solution of the linearization of the quadratic problem.
      epsilon: An epsilon scheduler or a float passed on to the linearization.
      old_transport_mass: Sum of the elements of the transport matrix at the
        previous iteration.

    Returns:
      Updated linear OT problem, a new local linearization of GW problem.
    """
    transport_mass = 1.0
    rescale_factor = 1.0
    unbalanced_correction = 0.0

    marginal_1 = transport.marginal(axis=1)
    marginal_2 = transport.marginal(axis=0)
    marginal_cost = self.marginal_dependent_cost(marginal_1, marginal_2)

    if not self.is_balanced:
      # Rescales transport for Unbalanced GW according to Sejourne et al (2021).
      transport_mass = jax.lax.stop_gradient(marginal_1.sum())
      rescale_factor = jnp.sqrt(old_transport_mass / transport_mass)
      unbalanced_correction = self.cost_unbalanced_correction(
          transport.matrix, marginal_1, marginal_2, epsilon, rescale_factor)
      # Updates epsilon for Unbalanced GW.
      epsilon = update_epsilon_unbalanced(epsilon, transport_mass)

    tmp = self.geom_xx.apply_cost(
        transport.matrix, axis=1, fn=self.quad_loss[0])
    tmp = self.geom_yy.apply_cost(tmp.T, axis=1, fn=self.quad_loss[1]).T

    cost_matrix = marginal_cost.cost_matrix - tmp + unbalanced_correction

    cost_matrix += self.fused_penalty * jnp.where(
        self.is_fused,
        0.0 if self.geom_xy is None else self.geom_xy.cost_matrix,
        0.0)

    cost_matrix *= rescale_factor

    geom = geometry.Geometry(cost_matrix=cost_matrix, epsilon=epsilon)
    return problems.LinearProblem(
        geom, self.a, self.b, tau_a=self.tau_a, tau_b=self.tau_b)

  def update_lr_linearization(
      self,
      lr_sink: sinkhorn_lr.LRSinkhornOutput) -> problems.LinearProblem:
    """Updates a Quad problem linearization using a LR Sinkhorn."""
    return problems.LinearProblem(
        self.update_lr_geom(lr_sink),
        self.a,
        self.b,
        tau_a=self.tau_a,
        tau_b=self.tau_b)


def update_epsilon_unbalanced(epsilon, transport_mass):
  updated_epsilon = epsilon_scheduler.Epsilon.make(epsilon)
  updated_epsilon._scale = updated_epsilon._scale * transport_mass
  return updated_epsilon


def make(*args,
         a: Optional[jnp.ndarray] = None,
         b: Optional[jnp.ndarray] = None,
         tau_a: float = 1.0,
         tau_b: float = 1.0,
         objective: Optional[str] = None,
         gw_unbalanced_correction: Optional[bool] = True,
         fused_penalty: Optional[float] = None,
         **kwargs):
  """Makes a problem from arrays, assuming PointCloud geometries."""
  if isinstance(args[0], (jnp.ndarray, np.ndarray)):
    x = args[0]
    y = args[1] if len(args) > 1 else args[0]
    if ((objective == 'linear') or
        (objective is None and x.shape[1] == y.shape[1])):
      geom_xy = pointcloud.PointCloud(x, y, **kwargs)
      return problems.LinearProblem(geom_xy, a=a, b=b, tau_a=tau_a, tau_b=tau_b)
    elif ((objective == 'quadratic') or
          (objective is None and x.shape[1] != y.shape[1])):
      geom_xx = pointcloud.PointCloud(x, x, **kwargs)
      geom_yy = pointcloud.PointCloud(y, y, **kwargs)
      return QuadraticProblem(geom_xx=geom_xx, geom_yy=geom_yy,
                              geom_xy=None,
                              a=a, b=b, tau_a=tau_a, tau_b=tau_b,
                              gw_unbalanced_correction=gw_unbalanced_correction)
    elif objective == 'fused':
      geom_xx = pointcloud.PointCloud(x, x, **kwargs)
      geom_yy = pointcloud.PointCloud(y, y, **kwargs)
      geom_xy = pointcloud.PointCloud(x, y, **kwargs)
      return QuadraticProblem(geom_xx=geom_xx, geom_yy=geom_yy, geom_xy=geom_xy,
                              fused_penalty=fused_penalty,
                              a=a, b=b, tau_a=tau_a, tau_b=tau_b,
                              gw_unbalanced_correction=gw_unbalanced_correction)
    else:
      raise ValueError(f'Unknown transport problem `{objective}`')
  elif isinstance(args[0], geometry.Geometry):
    cls = problems.LinearProblem if len(args) == 1 else QuadraticProblem
    return cls(*args, a=a, b=b, tau_a=tau_a, tau_b=tau_b)
  elif isinstance(args[0], (problems.LinearProblem, QuadraticProblem)):
    return args[0]
  else:
    raise ValueError('Cannot instantiate a transport problem.')



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some utility functions for transport computation.

This module is primarily made for new users who are looking for one-liners.
For instance, solving the transport between two point clouds.

>>> x = jax.random.uniform((34, 3))
>>> y = jax.random.uniform((34, 3)) + 1
>>> ot = ott.transport.solve(x, y)
>>> Tz = ot.apply(z)

Even if the transport.solve sole function can support many complex use cases, we
suggest more advanced users to instantiate directly their problem (see
ott.core.problems) and their solvers (see ott.core.sinkhorn and
ott.core.gromov_wasserstein) for better control over the parameters.
"""

from typing import Any, NamedTuple
import jax.numpy as jnp
from ott.core import gromov_wasserstein
from ott.core import problems
from ott.core import quad_problems
from ott.core import sinkhorn


class Transport(NamedTuple):
  """Implements a core.problems.Transport interface to transport solutions."""
  problem: Any = None
  solver_output: Any = None

  @property
  def linear(self):
    return isinstance(self.problem, problems.LinearProblem)

  @property
  def geom(self):
    return self.problem.geom if self.linear else self.solver_output.geom

  @property
  def a(self):
    return self.problem.a

  @property
  def b(self):
    return self.problem.b

  @property
  def linear_output(self):
    out = self.solver_output
    return out if self.linear else out.linear_state

  @property
  def matrix(self):
    return self.linear_output.matrix

  def apply(self, inputs: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    return self.linear_output.apply(inputs, axis)

  def marginal(self, axis: int = 0) -> jnp.ndarray:
    return self.linear_output.marginal(axis)


def solve(*args, a=None, b=None, objective=None, **kwargs) -> Transport:
  """Generic interface to transport problem.

  The geometries can be passed as arrays, geometry.Geometry or directly as a
  problem. The solver is passed via kwargs.

  Args:
    *args: can be either a single argument, the geometry.Geometry instance, or
      for convenience only two jnp.ndarray<float> corresponding to two point
      clouds. In that case the regularization parameter epsilon must be set in
      the kwargs.
    a: the weights of the source. Uniform by default.
    b: the weights of the target. Uniform by default.
    objective: Optional[str], 'linear', 'quadratic', 'fused' or None. None
      means that the objective will be chosen based on the dimensionalities
      of the arrays.
    **kwargs: the keyword arguments passed to the point clouds and/or the
      solvers.

  Returns:
    A Transport object.
  """
  tau_a, tau_b = kwargs.get('tau_a', 1.0), kwargs.get('tau_b', 1.0)
  gw_unbalanced_correction = kwargs.pop('gw_unbalanced_correction', True)
  fused_penalty = kwargs.pop('fused_penalty', None)
  eps_keys = ['epsilon', 'init', 'target', 'decay']
  pb_kwargs = {k: v for k, v in kwargs.items() if k in eps_keys}
  pb = quad_problems.make(
      *args,
      objective=objective,
      a=a,
      b=b,
      tau_a=tau_a,
      tau_b=tau_b,
      gw_unbalanced_correction=gw_unbalanced_correction,
      fused_penalty=fused_penalty,
      **pb_kwargs)
  linear = isinstance(pb, problems.LinearProblem)
  solver_fn = sinkhorn.make if linear else gromov_wasserstein.make
  geom_keys = ['cost_fn', 'power', 'online']
  remove_keys = geom_keys + eps_keys if linear else geom_keys
  for key in remove_keys:
    kwargs.pop(key, None)
  solver = solver_fn(**kwargs)
  output = solver(pb)
  return Transport(pb, output)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax version of the regularised GW Solver (Peyre et al. 2016)."""
import functools
from typing import Any, Dict, NamedTuple, Optional, Union

import jax
import jax.numpy as jnp
from ott.core import fixed_point_loop
from ott.core import problems
from ott.core import quad_problems
from ott.core import sinkhorn
from ott.core import sinkhorn_lr
from ott.geometry import epsilon_scheduler
from ott.geometry import geometry
from ott.geometry import low_rank
from ott.geometry import pointcloud


class GWOutput(NamedTuple):
  """Holds the output of the Gromov-Wasserstein solver.

  Attributes:
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    convergence: Bool convergence flag for the outer GW iterations.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
    linear_state: State used to solve and store solutions to the local
      linearization of GW.
    geom: The geometry underlying the local linearization.
    old_transport_mass: Holds total mass of transport at previous iteration.
    transport: The transport matrix.
    reg_gw_cost: Regularized optimal transport cost of the linearization.
  """
  costs: Optional[jnp.ndarray] = None
  linear_convergence: Optional[jnp.ndarray] = None
  convergence: bool = False
  errors: Optional[jnp.ndarray] = None
  linear_state: Any = None
  geom: geometry.Geometry = None
  # Intermediate values.
  old_transport_mass: Optional[float] = 1.0

  def set(self, **kwargs) -> 'GWOutput':
    """Returns a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  @property
  def matrix(self):
    """Transport matrix."""
    rescale_factor = jnp.sqrt(
        self.old_transport_mass / self.linear_state.transport_mass())
    return self.linear_state.matrix * rescale_factor

  @property
  def reg_gw_cost(self):
    return self.linear_state.reg_ot_cost


class GWState(NamedTuple):
  """Holds the state of the Gromov-Wasserstein solver.

  Attributes:
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
    linear_state: State used to solve and store solutions to the local
      linearization of GW.
    linear_pb: Local linearization of the quadratic GW problem.
    old_transport_mass: Intermediary value of the mass of the transport matrix.
  """
  costs: Optional[jnp.ndarray] = None
  linear_convergence: Optional[jnp.ndarray] = None
  errors: Optional[jnp.ndarray] = None
  linear_state: Any = None
  linear_pb: Optional[problems.LinearProblem] = None
  # Intermediate values.
  old_transport_mass: Optional[float] = 1.0

  def set(self, **kwargs) -> 'GWState':
    """Returns a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  def update(self, iteration: int, linear_sol, linear_pb, store_errors: bool,
             old_transport_mass: float):
    costs = self.costs.at[iteration].set(linear_sol.reg_ot_cost)
    errors = None
    if store_errors and self.errors is not None:
      errors = self.errors.at[iteration, :].set(linear_sol.errors)
    linear_convergence = self.linear_convergence.at[iteration].set(
        linear_sol.converged)
    return self.set(linear_state=linear_sol,
                    linear_pb=linear_pb,
                    costs=costs,
                    linear_convergence=linear_convergence,
                    errors=errors,
                    old_transport_mass=old_transport_mass)


@jax.tree_util.register_pytree_node_class
class GromovWasserstein:
  """A Gromov Wasserstein solver."""

  def __init__(self,
               epsilon: Optional[float] = None,
               rank: int = -1,
               linear_ot_solver: Any = None,
               min_iterations: int = 5,
               max_iterations: int = 50,
               threshold: float = 1e-3,
               jit: bool = True,
               store_inner_errors: bool = False,
               **kwargs):
    default_epsilon = 1.0
    # Set epsilon value to default if needed, but keep track of whether None was
    # passed to handle the case where a linear_ot_solver is passed or not.
    self.epsilon = epsilon if epsilon is not None else default_epsilon
    self.rank = rank
    self.linear_ot_solver = linear_ot_solver
    if self.linear_ot_solver is None:
      # Detect if user requests low-rank solver. In that case the
      # default_epsilon makes little sense, since it was designed for GW.
      if self.is_low_rank:
        if epsilon is None:
          # Use default entropic regularization in LRSinkhorn if None was passed
          self.linear_ot_solver = sinkhorn_lr.LRSinkhorn(
            rank=self.rank, **kwargs)
        else:
          # If epsilon is passed, use it to replace the default LRSinkhorn value
          self.linear_ot_solver = sinkhorn_lr.LRSinkhorn(
            rank=self.rank,
            epsilon=self.epsilon, **kwargs)
      else:
        # When using Entropic GW, epsilon is not handled inside Sinkhorn, 
        # but rather added back to the Geometry object reinstantiated 
        # when linearizing the problem. Therefore no need to pass it to solver.
        self.linear_ot_solver = sinkhorn.Sinkhorn(**kwargs)

    self.min_iterations = min_iterations
    self.max_iterations = max_iterations
    self.threshold = threshold
    self.jit = jit
    self.store_inner_errors = store_inner_errors
    self._kwargs = kwargs

  @property
  def is_low_rank(self):
    return self.rank > 0

  def tree_flatten(self):
    return ([self.epsilon, self.rank,
             self.linear_ot_solver, self.threshold],
            dict(
                min_iterations=self.min_iterations,
                max_iterations=self.max_iterations,
                jit=self.jit,
                store_inner_errors=self.store_inner_errors,
                **self._kwargs))

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    return cls(
        epsilon=children[0],
        rank=children[1],
        linear_ot_solver=children[2],
        threshold=children[3],
        **aux_data)

  def not_converged(self, state, iteration):
    costs, i, tol = state.costs, iteration, self.threshold
    return jnp.logical_or(
        i <= 2,
        jnp.logical_and(
            jnp.isfinite(costs[i - 1]),
            jnp.logical_not(jnp.isclose(costs[i - 2], costs[i - 1], rtol=tol))))

  def __call__(self, prob: quad_problems.QuadraticProblem) -> GWOutput:
    # Consider converting problem first is using low-rank solver
    if self.is_low_rank:
      convert = (
          isinstance(prob.geom_xx, pointcloud.PointCloud) and
          prob.geom_xx.is_squared_euclidean and
          isinstance(prob.geom_yy, pointcloud.PointCloud) and
          prob.geom_yy.is_squared_euclidean
          )
      # Consider converting
      if convert:
        if not prob.is_fused or isinstance(prob.geom_xy,
                                           low_rank.LRCGeometry):
          prob.geom_xx = prob.geom_xx.to_LRCGeometry()
          prob.geom_yy = prob.geom_yy.to_LRCGeometry()
        else:
          if (isinstance(prob.geom_xy, pointcloud.PointCloud) and
              prob.geom_xy.is_squared_euclidean):
            prob.geom_xy = prob.geom_xy.to_LRCGeometry(prob.fused_penalty)
            prob.geom_xx = prob.geom_xx.to_LRCGeometry()
            prob.geom_yy = prob.geom_yy.to_LRCGeometry()

    # Possibly jit iteration functions and run. Closure on rank to
    # avoid jitting issues, since rank value will be used to branch between
    # a default entropic GW or a low-rank GW.
    iterations_fn = functools.partial(iterations, rank=self.rank)
    gromov_fn = jax.jit(iterations_fn) if self.jit else iterations_fn
    out = gromov_fn(self, prob)
    # TODO(lpapaxanthos): remove stop_gradient when using backprop
    if self.rank > 0:
      linearization = prob.update_lr_linearization(
          jax.lax.stop_gradient(out.linear_state))
    else:
      linearization = prob.update_linearization(
          jax.lax.stop_gradient(out.linear_state),
          self.epsilon,
          jax.lax.stop_gradient(out.old_transport_mass))
    linear_state = out.linear_state.set_cost(linearization, True, True)
    iteration = jnp.sum(out.costs != 0)
    convergence = jnp.logical_not(self.not_converged(out, iteration))
    return out.set(linear_state=linear_state,
                   convergence=convergence)

  def init_state(self, prob: quad_problems.QuadraticProblem,
                 rank: int) -> GWState:
    """Initializes the state of the Gromov-Wasserstein iterations."""
    if rank > 0:
      linearization = prob.init_lr_linearization(rank)
    else:
      linearization = prob.init_linearization(self.epsilon)

    linear_state = self.linear_ot_solver(linearization)
    num_iter = self.max_iterations
    transport_mass = prob.init_transport_mass()
    if self.store_inner_errors:
      errors = -jnp.ones((num_iter, self.linear_ot_solver.outer_iterations))
    else:
      errors = None
    return GWState(-jnp.ones((num_iter,)), -jnp.ones((num_iter,)),
                   errors, linear_state, linearization, transport_mass)

  def output_from_state(self, state):
    """Create an output from a loop state.

    Arguments:
      state: A GWState.

    Returns:
      A GWOutput.
    """
    geom = state.linear_pb.geom
    return GWOutput(costs=state.costs,
                    linear_convergence=state.linear_convergence,
                    errors=state.errors,
                    linear_state=state.linear_state,
                    geom=geom,
                    old_transport_mass=state.old_transport_mass)


def iterations(solver: GromovWasserstein,
               prob: quad_problems.QuadraticProblem,
               rank: int) -> GWOutput:
  """A jittable Gromov-Wasserstein outer loop."""

  def cond_fn(iteration, constants, state):
    solver = constants
    return solver.not_converged(state, iteration)

  def body_fn(iteration, constants, state, compute_error):
    del compute_error  # Always assumed True for outer loop of GW.
    solver = constants
    if rank > 0:
      linear_pb = prob.update_lr_linearization(state.linear_state)
    else:
      linear_pb = prob.update_linearization(state.linear_state, solver.epsilon,
                                            state.old_transport_mass)

    out = solver.linear_ot_solver(linear_pb)
    old_transport_mass = jax.lax.stop_gradient(
        state.linear_state.transport_mass())
    return state.update(
        iteration,
        out,
        linear_pb,
        solver.store_inner_errors,
        old_transport_mass)

  state = fixed_point_loop.fixpoint_iter(
      cond_fn=cond_fn,
      body_fn=body_fn,
      min_iterations=solver.min_iterations,
      max_iterations=solver.max_iterations,
      inner_iterations=1,
      constants=solver,
      state=solver.init_state(prob, rank))

  return solver.output_from_state(state)


def make(
    epsilon: Union[epsilon_scheduler.Epsilon, float] = 1.,
    rank: int = -1,
    max_iterations: int = 50,
    jit: bool = False,
    warm_start: bool = True,
    store_inner_errors: bool = False,
    linear_ot_solver_kwargs: Optional[Dict[str, Any]] = None,
    threshold: float = 1e-2,
    min_iterations: int = 1,
    **kwargs) -> GromovWasserstein:
  """Creates a GromovWasserstein solver.

  Args:
    epsilon: a regularization parameter or a epsilon_scheduler.Epsilon object.
    rank: integer used to constrain the rank of GW solutions if >0.
    max_iterations: int32, the maximum number of outer iterations for
      Gromov Wasserstein.
    jit: bool, if True, jits the function.
    warm_start: deprecated.
    store_inner_errors: whether or not to return all the errors of the inner
      Sinkhorn iterations.
    linear_ot_solver_kwargs: Optionally a dictionary containing the keywords 
      arguments for the linear OT solver (e.g. sinkhorn)
    threshold: threshold (progress between two iterate costs) used to stop GW.
    min_iterations: see fixed_point_loop.
    **kwargs: additional kwargs for epsilon.

  Returns:
    A GromovWasserstein solver.
  """
  del warm_start
  if linear_ot_solver_kwargs is None:
    linear_ot_solver_kwargs = {}

  if rank == -1:
    sink = sinkhorn.make(**linear_ot_solver_kwargs)
  elif rank > 0:
    sink = sinkhorn_lr.make(**linear_ot_solver_kwargs)

  return GromovWasserstein(
      epsilon, rank, max_iterations=max_iterations,
      jit=jit, linear_ot_solver=sink, threshold=threshold,
      store_inner_errors=store_inner_errors,
      min_iterations=min_iterations, **kwargs)


def gromov_wasserstein(
    geom_xx: geometry.Geometry,
    geom_yy: geometry.Geometry,
    geom_xy: Optional[geometry.Geometry] = None,
    fused_penalty: Optional[float] = None,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    loss: Optional[str] = None,
    tau_a: Optional[float] = 1.0,
    tau_b: Optional[float] = 1.0,
    gw_unbalanced_correction: bool = True,
    **kwargs) -> GWOutput:
  """Wrapper to solve a Gromov Wasserstein problem.

  Wrapper that instantiates a quadratic problem (possibly with linear term
  if the problem is fused) and calls a solver to output a solution.

  Args:
    geom_xx: a Geometry object for the first view.
    geom_yy: a second Geometry object for the second view.
    geom_xy: a Geometry object representing the linear cost in FGW.
    fused_penalty: multiplier of the linear term in Fused Gromov Wasserstein,
      i.e. loss = quadratic_loss + fused_penalty * linear_loss. If geom_xy is
      None fused_penalty will be ignored, i.e. fused_penalty = 0
    a: jnp.ndarray<float>[num_a,] or jnp.ndarray<float>[batch,num_a] weights.
    b: jnp.ndarray<float>[num_b,] or jnp.ndarray<float>[batch,num_b] weights.
    loss: str, None defaults to the square Euclidean distance, can also
      receive 'kl' to define the GW loss.
    tau_a: float between 0 and 1.0, parameter that controls the strength of the
      KL divergence constraint between the weights and marginals of the
      transport for the first view. If set to 1.0, then it is equivalent to a
      hard constraint and if smaller to a softer constraint.
    tau_b: float between 0 and 1.0, parameter that controls the strength of the
      KL divergence constraint between the weights and marginals of the
      transport for the second view. If set to 1.0, then it is equivalent to a
      hard constraint and if smaller to a softer constraint.
    gw_unbalanced_correction: True (default) if the unbalanced version of
      Sejourne et al (Neurips 2021) is used, False if tau_a and tau_b
      only affect the inner Sinhkorn loop.
    **kwargs: keyword arguments to make.

  Returns:
    A GromovWassersteinState named tuple.
  """
  losses = {'kl': quad_problems.make_kl_loss}
  loss_fn = losses.get(loss, None)
  prob = quad_problems.QuadraticProblem(
      geom_xx,
      geom_yy,
      geom_xy=geom_xy,
      fused_penalty=fused_penalty,
      a=a,
      b=b,
      loss=(loss_fn() if loss_fn is not None else None),
      tau_a=tau_a,
      tau_b=tau_b,
      gw_unbalanced_correction=gw_unbalanced_correction)
  solver = make(**kwargs)
  return solver(prob)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes defining OT problem(s) (objective function + utilities)."""

from typing import Optional, Tuple
import jax
import jax.numpy as jnp
from ott.geometry import geometry


@jax.tree_util.register_pytree_node_class
class LinearProblem:
  """Holds the definition of a linear regularized OT problem and some tools."""

  def __init__(self,
               geom: geometry.Geometry,
               a: Optional[jnp.ndarray] = None,
               b: Optional[jnp.ndarray] = None,
               tau_a: float = 1.0,
               tau_b: float = 1.0):
    """Initializes the LinearProblem.

    min_P<C, P> - eps H(P), s.t P.1 = a, Pt.1 = b.

    Args:
      geom: the geometry.Geometry object defining the ground geometry / cost of
        the linear problem.
      a: jnp.ndarray[n] representing the first marginal. If None, it will be
        uniform.
      b: jnp.ndarray[n] representing the first marginal. If None, it will be
        uniform.
      tau_a: if lower that 1.0, defines how much unbalanced the problem is on
        the first marginal.
      tau_b: if lower that 1.0, defines how much unbalanced the problem is on
        the second marginal.
    """
    self.geom = geom
    self._a = a
    self._b = b
    self.tau_a = tau_a
    self.tau_b = tau_b

  def tree_flatten(self):
    return ([self.geom, self._a, self._b],
            {'tau_a': self.tau_a, 'tau_b': self.tau_b})

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    return cls(*children, **aux_data)

  @property
  def a(self):
    num_a = self.geom.shape[0]
    return jnp.ones((num_a,)) / num_a if self._a is None else self._a

  @property
  def b(self):
    num_b = self.geom.shape[1]
    return jnp.ones((num_b,)) / num_b if self._b is None else self._b

  @property
  def is_balanced(self):
    return self.tau_a == 1.0 and self.tau_b == 1.0

  @property
  def epsilon(self):
    return self.geom.epsilon

  def get_transport_functions(self, lse_mode: bool):
    """Instantiates useful functions for Sinkhorn depending on lse_mode."""
    geom = self.geom
    if lse_mode:
      marginal_a = lambda f, g: geom.marginal_from_potentials(f, g, 1)
      marginal_b = lambda f, g: geom.marginal_from_potentials(f, g, 0)
      app_transport = geom.apply_transport_from_potentials
    else:
      marginal_a = lambda f, g: geom.marginal_from_scalings(
          geom.scaling_from_potential(f), geom.scaling_from_potential(g), 1)
      marginal_b = lambda f, g: geom.marginal_from_scalings(
          geom.scaling_from_potential(f), geom.scaling_from_potential(g), 0)
      app_transport = lambda f, g, z, axis: geom.apply_transport_from_scalings(
          geom.scaling_from_potential(f),
          geom.scaling_from_potential(g), z, axis)
    return marginal_a, marginal_b, app_transport



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pytree for a normal distribution."""

import math
from typing import Optional

import jax
import jax.numpy as jnp
from ott.tools.gaussian_mixture import scale_tril

LOG2PI = math.log(2. * math.pi)


@jax.tree_util.register_pytree_node_class
class Gaussian:
  """PyTree for a normal distribution."""

  def __init__(self, loc: jnp.ndarray, scale: scale_tril.ScaleTriL):
    self._loc = loc
    self._scale = scale

  @classmethod
  def from_random(
      cls,
      key: jnp.ndarray,
      n_dimensions: int,
      stdev: float = 0.1,
      dtype: Optional[jnp.dtype] = None
  ) -> 'Gaussian':
    """Construct a random Gaussian.

    Args:
      key: jax.random seed
      n_dimensions: desired covariance dimensions
      stdev: standard deviation of loc and log eigenvalues
        (means for both are 0)
      dtype: data type

    Returns:
      A random Gaussian.
    """
    key, subkey0, subkey1 = jax.random.split(key, num=3)
    loc = jax.random.normal(
        key=subkey0, shape=(n_dimensions,), dtype=dtype) * stdev
    scale = scale_tril.ScaleTriL.from_random(
        key=subkey1, n_dimensions=n_dimensions, stdev=stdev, dtype=dtype)
    return cls(loc=loc, scale=scale)

  @classmethod
  def from_mean_and_cov(
      cls,
      mean: jnp.ndarray,
      cov: jnp.ndarray):
    """Construct a Gaussian from a mean and covariance."""
    scale = scale_tril.ScaleTriL.from_covariance(cov)
    return cls(loc=mean, scale=scale)

  @property
  def loc(self) -> jnp.ndarray:
    return self._loc

  @property
  def scale(self) -> scale_tril.ScaleTriL:
    return self._scale

  @property
  def n_dimensions(self) -> int:
    return self.loc.shape[-1]

  def covariance(self) -> jnp.ndarray:
    return self.scale.covariance()

  def to_z(self, x: jnp.ndarray) -> jnp.ndarray:
    return self.scale.centered_to_z(x_centered=x - self.loc)

  def from_z(self, z: jnp.ndarray) -> jnp.ndarray:
    return self.scale.z_to_centered(z=z) + self.loc

  def log_prob(
      self,
      x: jnp.ndarray,  # (?, d)
  ) -> jnp.ndarray:  # (?, d)
    """Log probability for a gaussian with a diagonal covariance."""
    d = x.shape[-1]
    z = self.to_z(x)
    log_det = self.scale.log_det_covariance()
    return (-0.5 * (d * LOG2PI +
                    log_det[None] +
                    jnp.sum(z ** 2., axis=-1)))  # (?, k)

  def sample(self, key: jnp.ndarray, size: int) -> jnp.ndarray:
    """Generate samples from the distribution."""
    std_samples_t = jax.random.normal(
        key=key, shape=(self.n_dimensions, size))
    return self.loc[None] + (
        jnp.swapaxes(jnp.matmul(self.scale.cholesky(), std_samples_t),
                     axis1=-2, axis2=-1))

  def w2_dist(self, other: 'Gaussian') -> jnp.ndarray:
    r"""Wasserstein distance W_2^2 to another Gaussian.

    W_2^2 = ||\mu_0-\mu_1||^2 +
       \text{trace} ( (\Lambda_0^\frac{1}{2} - \Lambda_1^\frac{1}{2})^2 )

    Args:
      other: other Gaussian

    Returns:
      The W_2^2 distance between self and other
    """
    delta_mean = jnp.sum((self.loc - other.loc) ** 2., axis=-1)
    delta_sigma = self.scale.w2_dist(other.scale)
    return delta_mean + delta_sigma

  def transport(
      self,
      dest: 'Gaussian',
      points: jnp.ndarray) -> jnp.ndarray:
    return self.scale.transport(
        dest_scale=dest.scale,
        points=points - self.loc[None]) + dest.loc[None]

  def tree_flatten(self):
    children = (self.loc, self.scale)
    aux_data = {}
    return children, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    return cls(*children, **aux_data)

  def __hash__(self):
    return jax.tree_util.tree_flatten(self).__hash__()

  def __eq__(self, other):
    return jax.tree_util.tree_flatten(self) == jax.tree_util.tree_flatten(other)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A class describing low-rank geometries."""
import jax
import jax.numpy as jnp
from ott.geometry import geometry


@jax.tree_util.register_pytree_node_class
class LRCGeometry(geometry.Geometry):
  r"""Low-rank Cost Geometry defined by two factors.
  """

  def __init__(self,
               cost_1: jnp.ndarray,
               cost_2: jnp.ndarray,
               bias: float = 0.0,
               **kwargs
               ):
    r"""Initializes a geometry by passing it low-rank factors.

    Args:
      cost_1: jnp.ndarray<float>[num_a, r]
      cost_2: jnp.ndarray<float>[num_b, r]
      bias: constant added to entire cost matrix.
      **kwargs: additional kwargs to Geometry
    """
    assert cost_1.shape[1] == cost_2.shape[1]
    self.cost_1 = cost_1
    self.cost_2 = cost_2
    self.bias = bias
    self._kwargs = kwargs

    super().__init__(**kwargs)

  @property
  def cost_rank(self):
    return self.cost_1.shape[1]

  @property
  def cost_matrix(self):
    """Returns cost matrix if requested."""
    return jnp.matmul(self.cost_1, self.cost_2.T) + self.bias

  @property
  def shape(self):
    return (self.cost_1.shape[0], self.cost_2.shape[0])

  @property
  def is_symmetric(self):
    return (self.cost_1.shape[0] == self.cost_2.shape[0] and
            jnp.all(self.cost_1 == self.cost_2))

  def apply_square_cost(self, arr: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Applies elementwise-square of cost matrix to array (vector or matrix)."""
    (n, m), r = self.shape, self.cost_rank
    # When applying square of a LRCgeometry, one can either elementwise square
    # the cost matrix, or instantiate an augmented (rank^2) LRCGeometry
    # and apply it. First is O(nm), the other is O((n+m)r^2).
    if n * m < (n + m) * r**2:  #  better use regular apply
      return super().apply_square_cost(arr, axis)
    else:
      new_cost_1 = self.cost_1[:, :, None] * self.cost_1[:, None, :]
      new_cost_2 = self.cost_2[:, :, None] * self.cost_2[:, None, :]
      return LRCGeometry(
          cost_1=new_cost_1.reshape((n, r**2)),
          cost_2=new_cost_2.reshape((m, r**2))).apply_cost(arr, axis)

  def _apply_cost_to_vec(self,
                         vec: jnp.ndarray,
                         axis: int = 0,
                         fn=None) -> jnp.ndarray:
    """Applies [num_a, num_b] fn(cost) (or transpose) to vector.

    Args:
      vec: jnp.ndarray [num_a,] ([num_b,] if axis=1) vector
      axis: axis on which the reduction is done.
      fn: function optionally applied to cost matrix element-wise, before the
        doc product

    Returns:
      A jnp.ndarray corresponding to cost x vector
    """
    def efficient_apply(vec, axis, fn):
      c1 = self.cost_1 if axis == 1 else self.cost_2
      c2 = self.cost_2 if axis == 1 else self.cost_1
      c2 = fn(c2) if fn is not None else c2
      bias = fn(self.bias) if fn is not None else self.bias
      out = jnp.dot(c1, jnp.dot(c2.T, vec))
      return out + bias * jnp.sum(vec) * jnp.ones_like(out)

    return jnp.where(fn is None or geometry.is_linear(fn),
                     efficient_apply(vec, axis, fn),
                     super()._apply_cost_to_vec(vec, axis, fn))

  def apply_cost_1(self, vec, axis=0):
    return jnp.dot(self.cost_1 if axis == 0 else self.cost_1.T, vec)

  def apply_cost_2(self, vec, axis=0):
    return jnp.dot(self.cost_2 if axis == 0 else self.cost_2.T, vec)

  def tree_flatten(self):
    return (self.cost_1, self.cost_2, self._kwargs), None

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del aux_data
    return cls(*children[:-1], **children[-1])


def add_lrc_geom(geom1: LRCGeometry, geom2: LRCGeometry):
  """Add geometry in geom1 to that in geom2, keeping other geom1 params."""
  return LRCGeometry(
      cost_1=jnp.concatenate((geom1.cost_1, geom2.cost_1), axis=1),
      cost_2=jnp.concatenate((geom1.cost_2, geom2.cost_2), axis=1),
      **geom1._kwargs)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pytree containing parameters for a pair of coupled Gaussian mixture models.
"""
import jax
import jax.numpy as jnp
from ott.core import sinkhorn
from ott.geometry import costs
from ott.geometry import geometry
from ott.geometry import pointcloud

from ott.tools.gaussian_mixture import gaussian_mixture


@jax.tree_util.register_pytree_node_class
class GaussianMixturePair:
  """Pytree for a coupled pair of Gaussian mixture models.

  Includes methods used in estimating an optimal pairing between GMM components
  using the Wasserstein-like method described in
  https://arxiv.org/abs/1907.05254 as well as generalization that allows for
  the re-weighting of components.

  The Delon & Desolneux paper above proposes fitting a pair of GMMs to a pair
  of point clouds in such a way that the sum of the log likelihood of the
  points minus a weighted penalty involving a Wasserstein-like distance between
  the GMMs. Their proposed algorithm involves using EM in which a balanced
  Sinkhorn algorithm is used to estimate a coupling between the GMMs at each
  step of EM.

  Our generalization of this algorithm allows for a mismatch between the
  marginals of the coupling and the GMM component weights. This mismatch can be
  interpreted as components being re-weighted rather than being transported.
  We penalize reweighting with a generalized KL-divergence penalty, and we give
  the option to use the unbalanced Sinkhorn algorithm rather than the balanced
  to compute the divergence between GMMs.
  """

  def __init__(
      self,
      gmm0: gaussian_mixture.GaussianMixture,
      gmm1: gaussian_mixture.GaussianMixture,
      epsilon: float = 1.e-2,
      tau: float = 1.,
      lock_gmm1: bool = False,
    ):
    """Constructor.

    When fitting a pair of coupled GMMs with *no* reweighting of components
    using the algorithm in https://arxiv.org/abs/1907.05254, set tau = 1. The
    coupling between components will be determined via the balanced Sinkhorn
    algorithm.

    When fitting a pair of coupled GMMs in which reweighting of components is
    allowed, set tau to a value in (0, 1). The resulting coupling will penalize
    the generalized KL divergence between the coupling's marginals and the GMM
    component weights with a weight of rho = epsilon tau / (1 - tau).

    Args:
      gmm0: first GMM in the pair
      gmm1: second GMM in the pair
      epsilon: regularization weight to use for the Sinkhorn algorithm
      tau: encodes the weight, rho, to use for the generalized KL divergence
        between the coupling's marginals and GMM component weights as
        rho = epsilon tau / (1 - tau)
      lock_gmm1: indicates whether the parameters of gmm1 should be modified
        during optimization
    """
    self._gmm0 = gmm0
    self._gmm1 = gmm1
    self._epsilon = epsilon
    self._tau = tau
    self._lock_gmm1 = lock_gmm1

  @property
  def dtype(self):
    return self.gmm0.dtype

  @property
  def gmm0(self):
    return self._gmm0

  @property
  def gmm1(self):
    return self._gmm1

  @property
  def epsilon(self):
    return self._epsilon

  @property
  def tau(self):
    return self._tau

  @property
  def rho(self):
    return self.epsilon * self.tau / (1. - self.tau)

  @property
  def lock_gmm1(self):
    return self._lock_gmm1

  def get_bures_geometry(self) -> pointcloud.PointCloud:
    """Get a Bures Geometry for the two GMMs."""
    mean0 = self.gmm0.loc
    dimension = mean0.shape[-1]
    cov0 = self.gmm0.covariance
    cov0 = cov0.reshape(cov0.shape[:-2] + (dimension * dimension,))
    x = jnp.concatenate([mean0, cov0], axis=-1)
    mean1 = self.gmm1.loc
    cov1 = self.gmm1.covariance
    cov1 = cov1.reshape(cov1.shape[:-2] + (dimension * dimension,))
    y = jnp.concatenate([mean1, cov1], axis=-1)
    return pointcloud.PointCloud(
        x=x, y=y,
        cost_fn=costs.Bures(dimension=dimension),
        epsilon=self.epsilon)

  def get_cost_matrix(self) -> jnp.ndarray:
    """Get matrix of W2^2 costs between all pairs of (gmm0, gmm1) components."""
    return self.get_bures_geometry().cost_matrix

  def get_sinkhorn(self, cost_matrix: jnp.ndarray) -> sinkhorn.SinkhornOutput:
    """Get the output of ott.sinkhorn's method for a given cost matrix."""
    # We use a Geometry here rather than the PointCloud created in
    # in get_bures_geometry to avoid recomputing the cost matrix, since
    # the cost matrix is quite expensive
    geom = geometry.Geometry(cost_matrix=cost_matrix, epsilon=self.epsilon)
    return sinkhorn.sinkhorn(
        geom,
        a=self.gmm0.component_weights,
        b=self.gmm1.component_weights,
        tau_a=self.tau, tau_b=self.tau)

  def get_normalized_sinkhorn_coupling(
      self,
      sinkhorn_output: sinkhorn.SinkhornOutput,
  ) -> jnp.ndarray:
    """Get the normalized coupling matrix for the specified Sinkhorn output.

    Args:
      sinkhorn_output: Sinkhorn algorithm output as returned by get_sinkhorn()

    Returns:
      A coupling matrix that tells how much of the mass of each component of
      gmm0 is mapped to each component of gmm1.
    """
    return sinkhorn_output.matrix / jnp.sum(sinkhorn_output.matrix)

  def tree_flatten(self):
    """Method used by jax.tree_util to flatten a GaussianMixturePair.

    We control the subset of parameters that we will optimize in fit_gmm_pair
    by selectively placing them in either children (the parameters to optimize)
    or aux_data (the parameters to leave alone).

    Returns:
      A tuple of child pytrees and a dict of auxiliary data.
    """
    children = [self.gmm0]
    aux_data = {'epsilon': self.epsilon,
                'tau': self.tau,
                'lock_gmm1': self.lock_gmm1}
    if self.lock_gmm1:
      aux_data['gmm1'] = self.gmm1
    else:
      children.append(self.gmm1)
    return tuple(children), aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    """Method used by jax.tree_util to unflatten a GaussianMixturePair.

    tree_flatten controls which parameters get optimized by placing them in
    either children or aux_data; here we invert the process.

    Args:
      aux_data: auxiliary data that is passed to the constructor as kwargs
      children: child pytrees passed to the constructor as args

    Returns:
      A GaussianMixturePair.
    """
    children = list(children)
    if 'gmm1' in aux_data:
      gmm1 = aux_data.pop('gmm1')
      children.insert(1, gmm1)
    return cls(*children, **aux_data)

  def __repr__(self):
    class_name = type(self).__name__
    children, aux = self.tree_flatten()
    return '{}({})'.format(
        class_name, ', '.join([repr(c) for c in children] +
                              [f'{k}: {repr(v)}' for k, v in aux.items()]))

  def __hash__(self):
    return jax.tree_util.tree_flatten(self).__hash__()

  def __eq__(self, other):
    return jax.tree_util.tree_flatten(self) == jax.tree_util.tree_flatten(other)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for Anderson acceleration."""
from typing import Any
import jax
import jax.numpy as jnp

from ott.core import dataclasses

SinkhornState = Any


@dataclasses.register_pytree_node
class AndersonAcceleration:
  """Implements Anderson acceleration for Sinkhorn."""

  memory: int = 2  # Number of iterates considered to form interpolation.
  refresh_every: int = 1  # Recompute interpolation periodically.
  ridge_identity: float = 1e-2  # Ridge used in the linear system.

  def extrapolation(self, xs, fxs):
    """Computes Anderson extrapolation from past observations."""
    # Remove -inf values to instantiate quadratic problem. All others
    # remain since they might be caused by a valid issue.
    fxs_clean = jnp.nan_to_num(fxs, nan=jnp.nan, posinf=jnp.inf, neginf=0.0)
    xs_clean = jnp.nan_to_num(xs, nan=jnp.nan, posinf=jnp.inf, neginf=0.0)
    residuals = fxs_clean - xs_clean
    gram_matrix = jnp.matmul(residuals.T, residuals)
    gram_matrix /= jnp.linalg.norm(gram_matrix)

    # Solve linear system to obtain weights
    weights = jax.scipy.sparse.linalg.cg(
        gram_matrix + self.ridge_identity * jnp.eye(xs.shape[1]),
        jnp.ones(xs.shape[1]))[0]
    weights /= jnp.sum(weights)

    # Recover linear combination and return it with NaN (caused
    # by 0 weights leading to -jnp.inf potentials, mixed with weights
    # coefficiences of different signs), disambiguated to -inf.
    combination = jnp.sum(fxs * weights[None, :], axis=1)
    return jnp.where(jnp.isfinite(combination), combination, -jnp.inf)

  def update(self,
             state: SinkhornState,
             iteration: int,
             pb, lse_mode: bool):
    """Anderson acceleration update.

    When using Anderson acceleration, first update the dual variable f_u with
    previous updates (if iteration count sufficiently large), then record
    new iterations in array.

    Anderson acceleration always happens in potentials (not scalings) space,
    regardless of the lse_mode setting. If the iteration count is large
    enough the update below will output a potential variable.

    Args:
      state: A sinkhorn.SinkhornState
      iteration: int, the current iteration.
      pb: a problem.LinearProblem defining the OT problem.
      lse_mode: whether to compute in log-sum-exp or in scalings.

    Returns:
      A potential variable.
    """
    geom = pb.geom
    trigger_update = jnp.logical_and(iteration > self.memory,
                                     iteration % self.refresh_every == 0)
    fu = jnp.where(trigger_update,
                   self.extrapolation(state.old_fus, state.old_mapped_fus),
                   state.fu)
    # If the interpolation was triggered, we store it in memory
    # Otherwise we add the previous value (converting it to potential form if
    # it was initially stored in scaling form).
    old_fus = jnp.where(
        trigger_update,
        jnp.concatenate((state.old_fus[:, 1:], fu[:, None]), axis=1),
        jnp.concatenate(
            (state.old_fus[:, 1:],
             (fu if lse_mode else geom.potential_from_scaling(fu))[:, None]),
            axis=1))

    # If update was triggered, ensure a scaling is returned, since the result
    # from the extrapolation was outputted in potential form.
    fu = jnp.where(
        trigger_update,
        fu if lse_mode else geom.scaling_from_potential(fu),
        fu)
    return state.set(fu=fu, old_fus=old_fus)

  def init_maps(self, pb, state):
    """Initializes log matrix used in Anderson acceleration with nan values."""
    fus = jnp.ones((pb.geom.shape[0], self.memory)) * jnp.nan
    return state.set(old_fus=fus, old_mapped_fus=fus)

  def update_history(self, state, pb, lse_mode: bool):
    f = state.fu if lse_mode else pb.geom.potential_from_scaling(state.fu)
    mapped = jnp.concatenate((state.old_mapped_fus[:, 1:], f[:, None]), axis=1)
    return state.set(old_mapped_fus=mapped)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions related to momemtum."""

import jax.numpy as jnp
from ott.core import dataclasses


@dataclasses.register_pytree_node
class Momentum:
  """Momentum for Sinkhorn updates, either constant or adaptive."""

  start: int = 0
  value: float = 1.0
  inner_iterations: int = 1

  def weight(self, state, iteration):
    """Computes momentum term if needed, using previously seen errors."""
    return jnp.where(
        iteration >= jnp.where(self.start == 0, jnp.inf, self.start),
        self.at(state), self.value)

  def at(self, state):
    """Momentum formula, https://arxiv.org/pdf/2012.12562v1.pdf, p.7 and (5)."""
    idx = self.start // self.inner_iterations
    error_ratio = jnp.minimum(
        state.errors[idx - 1, -1] / state.errors[idx - 2, -1], 0.99)
    power = 1.0 / self.inner_iterations
    return 2.0 / (1.0 + jnp.sqrt(1.0 - error_ratio ** power))

  def __call__(self, weight, value, new_value, lse_mode=True):
    if lse_mode:
      value = jnp.where(jnp.isfinite(value), value, 0.0)
      return (1.0 - weight) * value  + weight * new_value
    else:
      value = jnp.where(value > 0.0, value, 1.0)
      return value**(1.0 - weight) * new_value ** weight



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax implementation of the Low-Rank Sinkhorn algorithm."""
from typing import Optional, NamedTuple, Tuple, Any

import jax
import jax.numpy as jnp
from ott.core import fixed_point_loop
from ott.core import problems
from ott.core import sinkhorn
from ott.geometry import geometry


class LRSinkhornState(NamedTuple):
  """Holds the state of the Low Rank Sinkhorn algorithm."""
  q: Optional[jnp.ndarray] = None
  r: Optional[jnp.ndarray] = None
  g: Optional[jnp.ndarray] = None
  costs: Optional[jnp.ndarray] = None

  def set(self, **kwargs) -> 'LRSinkhornState':
    """Returns a copy of self, with potential overwrites."""
    return self._replace(**kwargs)

  def reg_ot_cost(self, ot_prob, use_danskin=False):
    return compute_reg_ot_cost(self.q, self.r, self.g, ot_prob, use_danskin)

  def solution_error(self, ot_prob, norm_error: jnp.ndarray, lse_mode: bool
                     ) -> jnp.ndarray:
    return solution_error(self.q, self.r, ot_prob, norm_error, lse_mode)


def compute_reg_ot_cost(q, r, g, ot_prob, use_danskin=False):
  q = jax.lax.stop_gradient(q) if use_danskin else q
  r = jax.lax.stop_gradient(r) if use_danskin else r
  g = jax.lax.stop_gradient(g) if use_danskin else g
  return jnp.sum(ot_prob.geom.apply_cost(r, axis=1) * q * (1.0 / g)[None, :])


def solution_error(q, r, ot_prob,
                   norm_error: jnp.ndarray, lse_mode: bool) -> jnp.ndarray:

  """Computes solution error.

  Since only balanced case is available for LR, this is marginal deviation.

  Args:
    q: first factor of solution
    r: second factor of solution
    ot_prob: linear problem
    norm_error: int, p-norm used to compute error.
    lse_mode: True if log-sum-exp operations, False if kernel vector products.

  Returns:
    one or possibly many numbers quantifying deviation to true marginals.
  """
  del lse_mode
  norm_error = jnp.array(norm_error)
  # Update the error
  err = jnp.sum(
      jnp.abs(jnp.sum(q, axis=1) - ot_prob.a) **
      norm_error[:, jnp.newaxis], axis=1) ** (1.0 / norm_error)
  err += jnp.sum(
      jnp.abs(jnp.sum(r, axis=1) - ot_prob.b) **
      norm_error[:, jnp.newaxis], axis=1) ** (1.0 / norm_error)
  err += jnp.sum(
      jnp.abs(jnp.sum(q, axis=0) - jnp.sum(r, axis=0)) **
      norm_error[:, jnp.newaxis], axis=1) ** (1.0 / norm_error)

  return err


class LRSinkhornOutput(NamedTuple):
  """Implements the problems.Transport interface, for a LR Sinkhorn solution."""
  q: Optional[jnp.ndarray] = None
  r: Optional[jnp.ndarray] = None
  g: Optional[jnp.ndarray] = None
  costs: Optional[jnp.ndarray] = None
  reg_ot_cost: Optional[jnp.ndarray] = None
  ot_prob: Optional[problems.LinearProblem] = None

  def set(self, **kwargs) -> 'LRSinkhornOutput':
    """Returns a copy of self, with potential overwrites."""
    return self._replace(**kwargs)

  def set_cost(self, ot_prob, lse_mode, use_danskin) -> 'LRSinkhornOutput':
    del lse_mode
    return self.set(reg_ot_cost=self.compute_reg_ot_cost(
        ot_prob, use_danskin))

  def compute_reg_ot_cost(self, ot_prob, use_danskin):
    return compute_reg_ot_cost(self.q, self.r, self.g,
                               ot_prob, use_danskin)

  @property
  def linear(self):
    return isinstance(self.ot_prob, problems.LinearProblem)

  @property
  def geom(self):
    return self.ot_prob.geom

  @property
  def a(self):
    return self.ot_prob.a

  @property
  def b(self):
    return self.ot_prob.b

  @property
  def linear_output(self):
    return True

  @property
  def converged(self):
    if self.costs is None:
      return False
    return jnp.logical_and(jnp.sum(self.costs == -1) > 0,
                           jnp.sum(jnp.isnan(self.costs)) == 0)

  @property
  def matrix(self) -> jnp.ndarray:
    """Transport matrix if it can be instantiated."""
    return jnp.matmul(self.q * (1/self.g)[None, :], self.r.T)

  def apply(self, inputs: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Applies the transport to a ndarray; axis=1 for its transpose."""
    q, r = (self.q, self.r) if axis == 1 else (self.r, self.q)
    return jnp.dot(q, jnp.dot(r.T, inputs) / self.g)

  def marginal(self, axis: int) -> jnp.ndarray:
    length = self.q.shape[0] if axis == 0 else self.r.shape[0]
    return self.apply(jnp.ones(length,), axis=axis)

  def cost_at_geom(self, other_geom: geometry.Geometry):
    """Returns OT cost for matrix, evaluated at other cost matrix."""
    return jnp.sum(
        self.q * other_geom.apply_cost(self.r, axis=1) / self.g[None, :])

  def transport_mass(self) -> float:
    """Sum of transport matrix."""
    return self.marginal(0).sum()


@jax.tree_util.register_pytree_node_class
class LRSinkhorn(sinkhorn.Sinkhorn):
  r"""A Low-Rank Sinkhorn solver for linear reg-OT problems.

  A Low-Rank Sinkhorn solver takes a linear OT problem as an input, to return
  a LRSinkhornOutput object.

  The algorithm is described in:
  Low-Rank Sinkhorn Factorization, Scetbon-Cuturi-Peyre, ICML'21.
  http://proceedings.mlr.press/v139/scetbon21a/scetbon21a.pdf

  and the implementation contained here is adapted from that of:
  https://github.com/meyerscetbon/LOT

  The algorithm minimizes a non-convex problem. It therefore requires special
  care to initialization and convergence. Initialization is random by default,
  and convergence evaluated on successive evaluations of the objective. The
  algorithm is only provided for the balanced case.

  Attributes:
    rank: the rank constraint on the coupling to minimize the linear OT problem
    gamma: the (inverse of) gradient stepsize used by mirror descent.
    epsilon: entropic regularization added on top of low-rank problem.
    lse_mode: whether to run computations in lse or kernel mode. At this moment,
      only ``lse_mode=True`` is implemented.
    threshold: convergence threshold, used to quantify whether two successive
      evaluations of the objective are (relatively) close enough to terminate.
    norm_error: norm used to quantify feasibility (deviation to marginals).
    inner_iterations: number of inner iterations used by the algorithm before
      reevaluating progress.
    min_iterations: min number of iterations before evaluating objective.
    max_iterations: max number of iterations allowed.
    use_danskin: use Danskin theorem to evaluate gradient of objective w.r.t.
      input parameters. Only ``True`` handled at this moment.
    implicit_diff: whether to use implicit differentiation. Not implemented
      at this moment.
    jit: jit by default iterations loop.
    rng_key: seed of random numer generator to initialize the LR factors.
    kwargs_dys : keyword arguments passed onto dysktra_update.
  """

  def __init__(self,
               rank: int = 10,
               gamma: float = 1.0,
               epsilon: float = 1e-4,
               lse_mode: bool = True,
               threshold: float = 1e-3,
               norm_error: int = 1,
               inner_iterations: int = 1,
               min_iterations: int = 0,
               max_iterations: int = 2000,
               use_danskin: bool = True,
               implicit_diff: bool = False,
               jit: bool = True,
               rng_key: int = 0,
               kwargs_dys: Any = None):
    self.rank = rank
    self.gamma = gamma
    self.epsilon = epsilon
    self.lse_mode = lse_mode
    assert lse_mode, "Kernel mode not yet implemented for LRSinkhorn."
    self.threshold = threshold
    self.inner_iterations = inner_iterations
    self.min_iterations = min_iterations
    self.max_iterations = max_iterations
    self._norm_error = norm_error
    self.jit = jit
    self.use_danskin = use_danskin
    self.implicit_diff = implicit_diff
    assert not implicit_diff, "Implicit diff. not yet implemented for LRSink."
    self.rng_key = rng_key
    self.kwargs_dys = {} if kwargs_dys is None else kwargs_dys

  def __call__(self,
               ot_prob: problems.LinearProblem,
               init: Optional[Tuple[Optional[jnp.ndarray], ...]] = None
               ) -> LRSinkhornOutput:
    """Main interface to run LR sinkhorn."""
    init_q, init_r, init_g = (init if init is not None else (None, None, None))
    # Random initialization for q, r, g using rng_key
    rng = jax.random.split(jax.random.PRNGKey(self.rng_key), 3)
    a, b = ot_prob.a, ot_prob.b
    if init_g is None:
      init_g = jnp.abs(jax.random.uniform(rng[0], (self.rank,))) + 1
      init_g = init_g / jnp.sum(init_g)
    if init_q is None:
      init_q = jnp.abs(jax.random.normal(rng[1], (a.shape[0], self.rank)))
      init_q = init_q * (a / jnp.sum(init_q, axis=1))[:, None]
    if init_r is None:
      init_r = jnp.abs(jax.random.normal(rng[2], (b.shape[0], self.rank)))
      init_r = init_r * (b / jnp.sum(init_r, axis=1))[:, None]
    run_fn = run if not self.jit else jax.jit(run)
    return run_fn(ot_prob, self, (init_q, init_r, init_g))

  @property
  def norm_error(self):
    return (self._norm_error,)

  def not_converged(self, state, iteration):
    costs, i, tol = state.costs, iteration, self.threshold
    return jnp.logical_or(
        i <= 2,
        jnp.logical_and(
            jnp.isfinite(costs[i - 1]),
            jnp.logical_not(jnp.isclose(costs[i - 2], costs[i - 1], rtol=tol))))

  def lr_costs(self, ot_prob, state, iteration):
    c_q = ot_prob.geom.apply_cost(state.r, axis=1) / state.g[None, :]
    c_q += (self.epsilon - 1 / self.gamma) * jnp.log(state.q)
    c_r = ot_prob.geom.apply_cost(state.q) / state.g[None, :]
    c_r += (self.epsilon - 1 / self.gamma) *  jnp.log(state.r)
    diag_qcr = jnp.sum(state.q * ot_prob.geom.apply_cost(state.r, axis=1),
                       axis=0)
    h = diag_qcr / state.g ** 2 - (
      self.epsilon - 1 / self.gamma) * jnp.log(state.g)
    return c_q, c_r, h

  def dysktra_update(self, c_q, c_r, h, ot_prob, state, iteration,
                     min_entry_value=1e-6, tolerance=1e-4,
                     min_iter=0, inner_iter=10, max_iter=200):

    # shortcuts for problem's definition.
    r = self.rank
    n, m = ot_prob.geom.shape
    a, b = ot_prob.a, ot_prob.b

    h_old = h
    g1_old, g2_old = jnp.zeros(r), jnp.zeros(r)
    f1, f2 = jnp.zeros(n), jnp.zeros(m)

    w_gi, w_gp = jnp.zeros(r), jnp.zeros(r)
    w_q, w_r = jnp.zeros(r), jnp.zeros(r)
    err = jnp.inf
    state_inner = f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err
    constants = c_q, c_r, a, b

    def cond_fn(iteration, constants, state_inner):
      del iteration, constants
      err = state_inner[-1]
      return err > tolerance

    def _softm(f, g, c, axis):
      return jax.scipy.special.logsumexp(
          self.gamma * (f[:, None] + g[None, :] - c), axis=axis)

    def body_fn(iteration, constants, state_inner, compute_error):
      f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err = state_inner
      c_q, c_r, a, b = constants

      # First Projection
      f1 = (jnp.log(a) - _softm(f1, g1_old, c_q, 1)) / self.gamma + f1
      f2 = (jnp.log(b) - _softm(f2, g2_old, c_r, 1)) / self.gamma + f2

      h = h_old + w_gi
      h = jnp.maximum(jnp.log(min_entry_value) / self.gamma, h)
      w_gi += h_old - h
      h_old = h

      # Update couplings
      g_q = _softm(f1, g1_old, c_q, 0)
      g_r = _softm(f2, g2_old, c_r, 0)

      # Second Projection
      h = (1 / 3) * (h_old + w_gp + w_q + w_r)
      h += g_q / (3 * self.gamma)
      h += g_r / (3 * self.gamma)
      g1 = h + g1_old - g_q / self.gamma
      g2 = h + g2_old - g_r / self.gamma

      w_q = w_q + g1_old - g1
      w_r = w_r + g2_old - g2
      w_gp = h_old + w_gp - h

      q, r, _ = self.recompute_couplings(f1, g1, c_q, f2, g2, c_r, h)

      g1_old = g1
      g2_old = g2
      h_old = h

      err = jnp.where(
          jnp.logical_and(compute_error, iteration >= min_iter),
          solution_error(q, r, ot_prob, self.norm_error, self.lse_mode), err)[0]

      return f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err

    state_inner = fixed_point_loop.fixpoint_iter_backprop(
        cond_fn, body_fn, min_iter, max_iter,
        inner_iter, constants, state_inner)

    f1, f2, g1_old, g2_old, h_old, _, _, _, _, _ = state_inner

    q, r, g = self.recompute_couplings(f1, g1_old, c_q, f2, g2_old, c_r, h_old)
    return q, r, g

  def recompute_couplings(self, f1, g1, c_q, f2, g2, c_r, h):
    q = jnp.exp(self.gamma * (f1[:, None] + g1[None, :] - c_q))
    r = jnp.exp(self.gamma * (f2[:, None] + g2[None, :] - c_r))
    g = jnp.exp(self.gamma * h)
    return q, r, g

  def lse_step(self, ot_prob, state, iteration) -> LRSinkhornState:
    """LR Sinkhorn LSE update."""
    c_q, c_r, h = self.lr_costs(ot_prob, state, iteration)
    q, r, g = self.dysktra_update(
        c_q, c_r, h, ot_prob, state, iteration, **self.kwargs_dys)
    return state.set(q=q, g=g, r=r)

  def kernel_step(self, ot_prob, state, iteration) -> LRSinkhornState:
    """LR Sinkhorn multiplicative update."""
    # TODO(cuturi): kernel step not implemented.
    return state

  def one_iteration(self, ot_prob, state, iteration, compute_error):
    """Carries out one LR sinkhorn iteration.

    Depending on lse_mode, these iterations can be either in:
      - log-space for numerical stability.
      - scaling space, using standard kernel-vector multiply operations.

    Args:
      ot_prob: the transport problem definition
      state: LRSinkhornState named tuple.
      iteration: the current iteration of the Sinkhorn outer loop.
      compute_error: flag to indicate this iteration computes/stores an error

    Returns:
      The updated state.
    """
    if self.lse_mode:  # In lse_mode, run additive updates.
      state = self.lse_step(ot_prob, state, iteration)
    else:
      state = self.kernel_step(ot_prob, state, iteration)

    # re-computes error if compute_error is True, else set it to inf.
    cost = jnp.where(
        jnp.logical_and(compute_error, iteration >= self.min_iterations),
        state.reg_ot_cost(ot_prob), jnp.inf)
    costs = state.costs.at[iteration // self.inner_iterations].set(cost)
    return state.set(costs=costs)

  def init_state(self, ot_prob, init):
    """Returns the initial state of the loop."""
    q, r, g = init
    costs = -jnp.ones(self.outer_iterations)
    return LRSinkhornState(q=q, r=r, g=g, costs=costs)

  def output_from_state(self, ot_prob, state):
    """Creates an output from a loop state.

    Args:
      ot_prob: the transport problem.
      state: a LRSinkhornState.

    Returns:
      A LRSinkhornOutput.
    """
    return LRSinkhornOutput(
        q=state.q, r=state.r, g=state.g, ot_prob=ot_prob, costs=state.costs)


def run(ot_prob, solver, init) -> LRSinkhornOutput:
  """Run loop of the solver, outputting a state upgraded to an output."""
  out = sinkhorn.iterations(ot_prob, solver, init)
  out = out.set_cost(ot_prob, solver.lse_mode, solver.use_danskin)
  return out.set(ot_prob=ot_prob)

def make(
  rank: int = 10,
  gamma: float = 1.0,
  epsilon: float = 1e-4,
  lse_mode: bool = True,
  threshold: float = 1e-3,
  norm_error: int = 1,
  inner_iterations: int = 1,
  min_iterations: int = 0,
  max_iterations: int = 2000,
  use_danskin: bool = True,
  implicit_diff: bool = False,
  jit: bool = True,
  rng_key: int = 0,
  kwargs_dys: Any = None) -> LRSinkhorn:
  
  return LRSinkhorn(
    rank=rank,
    gamma=gamma,
    epsilon=epsilon,
    lse_mode=lse_mode,
    threshold=threshold,
    norm_error=norm_error,
    inner_iterations=inner_iterations,
    min_iterations=min_iterations,
    max_iterations=max_iterations,
    use_danskin=use_danskin,
    implicit_diff=implicit_diff,
    jit=jit,
    rng_key=rng_key,
    kwargs_dys=kwargs_dys)